#include "chomik_compiler.h"

static std::unique_ptr<LLVMContext> TheContext;
static std::unique_ptr<IRBuilder<>> Builder;
static std::unique_ptr<Module> TheModule;
static std::map<std::string, Value *> NamedValues;
static std::unique_ptr<FunctionPassManager> TheFPM;
static std::unique_ptr<LoopAnalysisManager> TheLAM;
static std::unique_ptr<FunctionAnalysisManager> TheFAM;
static std::unique_ptr<CGSCCAnalysisManager> TheCGAM;
static std::unique_ptr<ModuleAnalysisManager> TheMAM;
static std::unique_ptr<PassInstrumentationCallbacks> ThePIC;
static std::unique_ptr<StandardInstrumentations> TheSI;
static ExitOnError ExitOnErr;

Function* chomik_compiler::parser::TheFunction = nullptr;
Value* chomik_compiler::parser::TheChomikMachine=nullptr;

Value *LogErrorV(const char *Str)
{
    std::cerr << Str << "\n";
    return nullptr;
}

Value *chomik_compiler::AST::simple_value_float_signature_itemAST::codegen()
{
    return ConstantFP::get(*TheContext, APFloat(value));
}

Value *chomik_compiler::AST::identifier_name_itemAST::codegen()
{
    Type* retTy = Type::getInt8Ty(*TheContext)->getPointerTo();
    FunctionType* FuncTy = FunctionType::get(retTy, false);
    PointerType* PtrToFuncTy = PointerType::get(FuncTy, 0);


    //AllocaInst * a = Builder->CreateAlloca(Type::getInt8Ty(*TheContext)->getPointerTo(), nullptr, "identifier");

    Value * v = Builder->CreateGlobalString(StringRef(identifier.c_str()),"identifier");

    std::vector<Value *> t0;
    t0.push_back(v);
    ArrayRef<Value*> Args{t0};

    auto Callee = parser::TheFunction->getParent()->getOrInsertFunction("chomik_create_identifier_literal", Type::getInt32Ty(*TheContext)->getPointerTo());
    auto fun = dyn_cast<Constant>(Callee.getCallee());

    Function *dofoo = cast<Function>(fun);

    Value * c=Builder->CreateCall(dofoo, Args, "the_identifier");
    //Value * d=Builder->CreateStore(c, a);

    return c;
}

Value *chomik_compiler::AST::float_literal_name_itemAST::codegen()
{
    return ConstantFP::get(*TheContext, APFloat(value));
}

Value *chomik_compiler::AST::int_literal_name_itemAST::codegen()
{
    return LogErrorV("no idea how to do it");
}

Value *chomik_compiler::AST::string_literal_name_itemAST::codegen()
{
    Type* retTy = Type::getInt8Ty(*TheContext)->getPointerTo();
    FunctionType* FuncTy = FunctionType::get(retTy, false);
    PointerType* PtrToFuncTy = PointerType::get(FuncTy, 0);

    //AllocaInst * a = Builder->CreateAlloca(Type::getInt8Ty(*TheContext)->getPointerTo(), nullptr, "the_string_literal");


    auto Callee = parser::TheFunction->getParent()->getOrInsertFunction("chomik_create_string_literal", Type::getInt32Ty(*TheContext)->getPointerTo());
    auto fun = dyn_cast<Constant>(Callee.getCallee());

    Function *dofoo = cast<Function>(fun);


    Value * v = Builder->CreateGlobalString(StringRef(value.c_str()),"value_string");


    std::vector<Value *> t0;
    t0.push_back(v);
    ArrayRef<Value*> Args{t0};

    Value * c=Builder->CreateCall(dofoo, Args, "the_string_literal");
    //Value * d=Builder->CreateStore(c, a);

    return c;
}

Value *chomik_compiler::AST::nonempty_list_of_name_itemsAST::codegen()
{
    AllocaInst * a = Builder->CreateAlloca(Type::getInt8Ty(*TheContext)->getPointerTo(), nullptr, "the_list_of_name_items");

    return a;
}

Value* chomik_compiler::AST::variable_execution_statementAST::codegen()
{
    AllocaInst * a = Builder->CreateAlloca(Type::getInt8Ty(*TheContext)->getPointerTo(), nullptr, "the_generic_name");


    Type* retTy = Type::getInt8Ty(*TheContext)->getPointerTo();
    FunctionType* FuncTy = FunctionType::get(retTy, false);
    PointerType* PtrToFuncTy = PointerType::get(FuncTy, 0);

    auto Callee = parser::TheFunction->getParent()->getOrInsertFunction("chomik_create_generic_name", Type::getInt32Ty(*TheContext)->getPointerTo());
    auto fun = dyn_cast<Constant>(Callee.getCallee());

    Function *dofoo = cast<Function>(fun);

    std::vector<Value *> t0;
    t0.push_back(parser::TheChomikMachine);
    ArrayRef<Value*> Args{t0};

    Value * c=Builder->CreateCall(dofoo, Args, "the_generic_name");
    //Value * d=Builder->CreateStore(c, a);


    for (auto & ni: my_vector)
    {
        Type* retTy2 = Type::getVoidTy(*TheContext);
        FunctionType* FuncTy2 = FunctionType::get(retTy2, false);
        PointerType* PtrToFuncTy2 = PointerType::get(FuncTy2, 0);

        auto Callee2 = parser::TheFunction->getParent()->getOrInsertFunction("chomik_generic_name_add_name_item", Type::getVoidTy(*TheContext));
        auto fun2 = dyn_cast<Constant>(Callee2.getCallee());

        Function *dofoo2 = cast<Function>(fun2);

        Value * x = ni->codegen();

        std::vector<Value *> t1;

        t1.push_back(c);
        t1.push_back(x);
        ArrayRef<Value*> Args1{t1};

        Value * c2=Builder->CreateCall(dofoo2, Args1);
    }

    // execute the variable
    Type* retTy3 = Type::getVoidTy(*TheContext);
    FunctionType* FuncTy3 = FunctionType::get(retTy3, false);
    PointerType* PtrToFuncTy3 = PointerType::get(FuncTy3, 0);

    auto Callee3 = parser::TheFunction->getParent()->getOrInsertFunction("chomik_execute_variable_value", Type::getVoidTy(*TheContext));
    auto fun3 = dyn_cast<Constant>(Callee3.getCallee());

    Function *dofoo3 = cast<Function>(fun3);

    std::vector<Value *> t3;

    t3.push_back(parser::TheChomikMachine);
    t3.push_back(c);
    ArrayRef<Value*> Args3{t3};

    Value * c3=Builder->CreateCall(dofoo3, Args3);



    Type* retTy2 = Type::getVoidTy(*TheContext);
    FunctionType* FuncTy2 = FunctionType::get(retTy2, false);
    PointerType* PtrToFuncTy2 = PointerType::get(FuncTy2, 0);

    auto Callee2 = parser::TheFunction->getParent()->getOrInsertFunction("chomik_destroy_generic_name", Type::getVoidTy(*TheContext));
    auto fun2 = dyn_cast<Constant>(Callee2.getCallee());

    Function *dofoo2 = cast<Function>(fun2);


    std::vector<Value *> t1;

    t1.push_back(parser::TheChomikMachine);
    t1.push_back(c);
    ArrayRef<Value*> Args1{t1};

    Value * c2=Builder->CreateCall(dofoo2, Args1);


    return c2;
}

Function* chomik_compiler::parser::MainLoop()
{
    TheFunction = nullptr;

    FunctionType *FT =
    FunctionType::get(Type::getInt32Ty(*TheContext), false);

    Function *F =
    Function::Create(FT, Function::ExternalLinkage, "execute_chomik_code", TheModule.get());

    TheFunction = F;


    if (!TheFunction)
        return nullptr;

    BasicBlock *BB = BasicBlock::Create(*TheContext, "entry", TheFunction);
    Builder->SetInsertPoint(BB);


    AllocaInst * a = Builder->CreateAlloca(Type::getInt8Ty(*TheContext)->getPointerTo(), nullptr, "the_machine");

    // call chomik_create_machine()

    Type* retTy = Type::getInt8Ty(*TheContext)->getPointerTo();
    FunctionType* FuncTy = FunctionType::get(retTy, false);
    PointerType* PtrToFuncTy = PointerType::get(FuncTy, 0);

    auto Callee = TheFunction->getParent()->getOrInsertFunction("chomik_create_machine", Type::getInt32Ty(*TheContext)->getPointerTo());
    auto fun = dyn_cast<Constant>(Callee.getCallee());

    Function *dofoo = cast<Function>(fun);

    TheChomikMachine=Builder->CreateCall(dofoo, std::nullopt, "the_machine");


    //
    Builder->CreateStore(TheChomikMachine, a);

    do
    {
        token = get_next_token();
        if (!token)
            break;

        auto x = parse_statement(token);

        if (!x)
        {
            std::cerr << "error in parsing statement\n";
            return nullptr;
        }

        if (auto *xprim = x->codegen())
        {

        }
    }
    while (true);

    // call chomik_destroy_machine()

    auto Callee2 = TheFunction->getParent()->getOrInsertFunction("chomik_destroy_machine", Type::getVoidTy(*TheContext));
    auto fun2 = dyn_cast<Constant>(Callee2.getCallee());
    Function *dofoo2 = cast<Function>(fun2);

    //Value * cl = Builder->CreateLoad(Type::getInt8Ty(*TheContext)->getPointerTo(), TheChomikMachine, "the_machine");

    std::vector<Value *> t0;
    t0.push_back(TheChomikMachine);
    ArrayRef<Value*> Args{t0};

    Builder->CreateCall(dofoo2, Args, "");

    Value * RetVal = ConstantInt::get(*TheContext, APInt(32, 0, true));
    Builder->CreateRet(RetVal);

    verifyFunction(*TheFunction);

    return TheFunction;
}

static void InitializeModuleAndManagers() {
    // Open a new context and module.
    TheContext = std::make_unique<LLVMContext>();
    TheModule = std::make_unique<Module>("chomik project", *TheContext);

    // Create a new builder for the module.
    Builder = std::make_unique<IRBuilder<>>(*TheContext);

    TheFPM = std::make_unique<FunctionPassManager>();
    TheLAM = std::make_unique<LoopAnalysisManager>();
    TheFAM = std::make_unique<FunctionAnalysisManager>();
    TheCGAM = std::make_unique<CGSCCAnalysisManager>();
    TheMAM = std::make_unique<ModuleAnalysisManager>();
    ThePIC = std::make_unique<PassInstrumentationCallbacks>();
    TheSI = std::make_unique<StandardInstrumentations>(*TheContext,
                                                       /*DebugLogging*/ true);
    TheSI->registerCallbacks(*ThePIC, TheMAM.get());
}


int main(int argc, char * argv[])
{
    chomik::program pr;
    chomik_compiler::parser pa{pr};

    InitializeModuleAndManagers();

    if (pa.parse(argv[1])==-1)
    {
        std::cerr << "error\n";
        return -1;
    }
    TheModule->print(errs(), nullptr);

    InitializeAllTargetInfos();
    InitializeAllTargets();
    InitializeAllTargetMCs();
    InitializeAllAsmParsers();
    InitializeAllAsmPrinters();

    auto TargetTriple = sys::getDefaultTargetTriple();
    TheModule->setTargetTriple(TargetTriple);

    std::string Error;
    auto Target = TargetRegistry::lookupTarget(TargetTriple, Error);

    // Print an error and exit if we couldn't find the requested target.
    // This generally occurs if we've forgotten to initialise the
    // TargetRegistry or we have a bogus target triple.
    if (!Target) {
        errs() << Error;
        return 1;
    }

    auto CPU = "generic";
    auto Features = "";

    TargetOptions opt;
    auto TheTargetMachine = Target->createTargetMachine(TargetTriple, CPU, Features, opt, Reloc::PIC_);

    TheModule->setDataLayout(TheTargetMachine->createDataLayout());


    auto Filename = "../tests/a.o";
    std::error_code EC;
    raw_fd_ostream dest(Filename, EC, sys::fs::OF_None);

    if (EC) {
        errs() << "Could not open file: " << EC.message();
        return 1;
    }

    legacy::PassManager pass;
    auto FileType = CodeGenFileType::ObjectFile;

    if (TheTargetMachine->addPassesToEmitFile(pass, dest, nullptr, FileType)) {
        errs() << "TargetMachine can't emit a file of this type";
        return 1;
    }

    pass.run(*TheModule);
    dest.flush();

    outs() << "Wrote " << Filename << "\n";

    return 0;
}
