#include "chomik_compiler.h"


extern "C" void chomik_open_file(const char * f);
extern "C" void chomik_close_file();
extern "C" int yylex( YYSTYPE * yylval_param, void * yyscanner );
extern "C" int yylineno();

int chomik_compiler::parser::token;
YYSTYPE chomik_compiler::parser::lvalue;


int chomik_compiler::parser::get_next_token()
{
    return yylex(&lvalue, nullptr);
}

int chomik_compiler::parser::get_line_number()
{
    return yylineno();
}



std::unique_ptr<chomik_compiler::AST::name_item_placeholderAST> chomik_compiler::parser::parse_name_item_placeholder(int first_token)
{
    return LogError<chomik_compiler::AST::name_item_placeholderAST>("failed to parse a name item placeholder");
}

std::unique_ptr<chomik_compiler::AST::code_literal_name_itemAST> chomik_compiler::parser::parse_code_literal_name_item(int first_token)
{
    return LogError<chomik_compiler::AST::code_literal_name_itemAST>("failed to parse a name item code literal");
}

std::unique_ptr<chomik_compiler::AST::nonempty_list_of_name_itemsAST> chomik_compiler::parser::parse_nonempty_list_of_name_items(int first_token)
{
    auto r = std::make_unique<chomik_compiler::AST::nonempty_list_of_name_itemsAST>();

    int token = first_token;
    while (token)
    {
        if (token == '>')
        {
            return std::move(r);
        }
        auto ni = parse_name_item(token);

        r->my_vector.push_back(std::move(ni));

        token = get_next_token();
    }

    return r;
}

std::unique_ptr<chomik_compiler::AST::name_itemAST> chomik_compiler::parser::parse_name_item(int first_token)
{
    if (first_token == T_IDENTIFIER)
    {
        auto x = std::make_unique<chomik_compiler::AST::identifier_name_itemAST>(lvalue.value_string);

        //outs() << lvalue.value_string;

        return x;
    }
    else
        if (first_token == '(')
        {
            return parse_name_item_placeholder(first_token);
        }
        else
            if (first_token == T_INT_LITERAL)
            {
                auto x = std::make_unique<chomik_compiler::AST::int_literal_name_itemAST>(lvalue.value_int);

                outs() << lvalue.value_int;
                return x;
            }
            else
                if (first_token == T_FLOAT_LITERAL)
                {
                    auto x = std::make_unique<chomik_compiler::AST::float_literal_name_itemAST>(lvalue.value_float);

                    outs() << lvalue.value_float;
                    return x;
                }
                else
                    if (first_token == T_STRING_LITERAL)
                    {
                        auto x = std::make_unique<chomik_compiler::AST::string_literal_name_itemAST>(lvalue.value_string);

                        outs() << lvalue.value_string;
                        return x;
                    }
                    else
                        if (first_token == '{')
                        {
                            return parse_code_literal_name_item(first_token);
                        }
                        else
                            if (first_token == '<')
                            {
                                auto x = parse_nonempty_list_of_name_items(first_token);
                                return LogError<chomik_compiler::AST::name_itemAST>("failed to parse a variable value name item");
                            }
                            return LogError<chomik_compiler::AST::name_itemAST>("failed to parse a name item");
}


std::unique_ptr<chomik_compiler::AST::assignment_statementAST> chomik_compiler::parser::parse_assignment_statement(int first_token)
{
    if (first_token == T_LET)
    {

    }
    return LogError<chomik_compiler::AST::assignment_statementAST>("failed to parse an assignment statement");;
}


std::unique_ptr<chomik_compiler::AST::variable_execution_statementAST> chomik_compiler::parser::parse_variable_execution_statement(int first_token)
{
    if (first_token == '<')
    {
        int token = get_next_token();
        auto c = parse_nonempty_list_of_name_items(token);

        if (!c)
        {
            return LogError<chomik_compiler::AST::variable_execution_statementAST>("failed to parse an execution statement");
        }

        auto r = std::make_unique<chomik_compiler::AST::variable_execution_statementAST>(std::move(c->my_vector));

        return std::move(r);
    }
    return LogError<chomik_compiler::AST::variable_execution_statementAST>("failed to parse an execution statement");
}

std::unique_ptr<chomik_compiler::AST::statementAST> chomik_compiler::parser::parse_statement(int first_token)
{
    int token;

    if (first_token == '<')
    {
        auto x = parse_variable_execution_statement(first_token);

        token = get_next_token();
        if (token == ';')
            return std::move(x);
    }
    else
    if (first_token == T_LET)
    {
        auto x = parse_assignment_statement(first_token);
        token = get_next_token();
        if (token == ';')
            return std::move(x);
    }

    return LogError<chomik_compiler::AST::statementAST>("failed to parse a statement");;
}

int chomik_compiler::parser::parse(const char * filename)
{
    chomik_open_file(filename);

    auto * x = MainLoop();
    if (!x)
        return -1;

    // no chomik_close_file is required,
    // it is done by the scanner
    return 0;
}
